class MetricsAliases:
    """Metrics key aliases.

    A class to provide human-readable labels for each key in a
    ``metrics_dict`` that we might want use when coloring a classification
    curve plot.
    """

    cbar_dict = {
        "tp": "True Positive (TP) Count",
        "fp": "False Positive (FP) Count",
        "fn": "False Negative (FN) Count",
        "tn": "True Negative (TN) Count",
        "tp_w": "Weighted True Positive (TP) Sum",
        "fp_w": "Weighted False Positive (FP) Sum",
        "fn_w": "Weighted False Negative (FN) Sum",
        "tn_w": "Weighted True Negative (TN) Sum",
        "precision": "Precision = TP/(TP + FP)",
        "f1": "F1 Score",
        "recall": "Recall = TP/(TP + FN)",
        "tpr": "Recall = TP/(TP + FN)",
        "fpr": "FPR = FP/(FP + TN)",
        "tpr_w": "Weighted Recall = TP/(TP + FN)",
        "fpr_w": "Weighted FPR = FP/(FP + TN)",
        "frac": "Fraction Flagged",
        "frac_w": "Weighted Fraction Flagged",
        "thresh": "Score Threshold Value",
    }



from typing import List, Optional, Tuple

import matplotlib
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from clscurves.config import MetricsAliases
from clscurves.covariance import CovarianceEllipseGenerator
from clscurves.utils import MetricsResult


class MetricsPlotter(MetricsAliases):
    """A helper class to provide methods shared by each metrics plotter.

    These methods streamline the process of making a single classification
    curve metrics plot, making a bootstrapped plot, and adding a confidence
    ellipse to a specified operating point.
    """

    def __init__(
        self,
        metrics: MetricsResult,
        score_is_probability: bool,
    ) -> None:
        self.metrics = metrics
        self.score_is_probability = score_is_probability

    def _add_op_ellipse(
        self,
        curves: pd.DataFrame,
        op_value: float,
        x_col: str,
        y_col: str,
        ax: plt.Axes,
        thresh_key: str = "thresh",
    ) -> None:
        """A helper function to add a confidence ellipse to an metrics plot
        given a threshold operating value.

        Parameters
        ----------
        curves : pd.DataFrame
            metrics.curves DataFrame.
        op_value : float
            Threshold operating value.
        x_col : str
            metrics.curves key used in plot x axis.
        y_col : str
            metrics.curves key used in plot y axis.
        ax : plt.Axes
            Matplotlib axis object.
        thresh_key : str
            metrics.curves key used for coloring (default: "thresh").
        """

        def find_op_point(df: pd.DataFrame) -> pd.Series:
            """Find all entries at or above the operating point threshold."""
            match = df[df[thresh_key] >= op_value]
            if not match.empty:
                return match.iloc[0]
            return df.iloc[0]

        # Get operating point coordinates for each bootstrapped sample
        op_points = curves.groupby("_bootstrap_sample", dropna=False).apply(
            find_op_point
        )
        op_data = op_points[[x_col, y_col]].values.T

        # Compute covariance ellipse and add to ax
        ceg = CovarianceEllipseGenerator(op_data)
        ceg.create_ellipse_patch(ax=ax, color="black")
        ceg.add_ellipse_center(ax=ax)

        # Add individual operating points
        ax.scatter(x=op_data[0], y=op_data[1], s=2, c="black", alpha=0.7, marker=".")

    def _make_plot(
        self,
        curves: pd.DataFrame,
        x_col: str,
        y_col: str,
        cmap: str,
        dpi: Optional[int],
        color_by: str,
        cbar_rng: Optional[List[float]],
        cbar_label: Optional[str],
        grid: bool,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """A helper function to create a base Matplotlib scatter plot figure
        for metrics-related plotting.
        """

        # Get non-bootstrapped data
        curves = curves.loc[lambda x: x["_bootstrap_sample"].isnull()]

        # Create figure
        if not ax:
            fig = plt.figure(figsize=(10, 7), dpi=dpi)
            ax = fig.add_subplot(1, 1, 1, aspect="equal")
            ax.grid(grid)
            ax.set_xlim(0, 1)
            ax.set_ylim(0, 1)

        # Make Color Bar
        color = curves[color_by]
        if cbar_rng is not None:
            [vmin, vmax] = cbar_rng
        else:
            sip = self.score_is_probability or color_by == "frac"
            vmin = 0.0 if sip else color.min()
            vmax = 1.0 if sip else color.max()
        norm = matplotlib.colors.Normalize(vmin, vmax)
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        sm.set_array(np.array([]))
        cbar = fig.colorbar(sm, ax=ax, ticks=np.linspace(vmin, vmax, 11))  # type: ignore
        default_cbar_label = (
            self.cbar_dict[color_by] if color_by in self.cbar_dict else "Value"
        )
        cbar_label = default_cbar_label if cbar_label is None else cbar_label
        cbar.set_label(cbar_label)

        # Make scatter plot
        print("Making scatter plot...")
        ax.scatter(
            curves[x_col],
            curves[y_col],
            s=100,
            c=color,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            marker=".",
            edgecolors="none",
            zorder=int(1e4),
        )

        return fig, ax  # type: ignore

    def _make_bootstrap_plot(
        self,
        curves: pd.DataFrame,
        x_col: str,
        y_col: str,
        cmap: str,
        dpi: Optional[int],
        color_by: str,
        cbar_rng: Optional[List[float]],
        cbar_label: Optional[str],
        grid: bool,
        alpha: float,
        bootstrap_color: str,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """A helper function to add faint bootstrapped reference curves to an
        metrics plot to visualize the confidence we have in the main metrics
        curve.
        """

        # Create figure
        fig = plt.figure(figsize=(10, 7), dpi=dpi)
        ax = fig.add_subplot(1, 1, 1, aspect="equal")
        ax.grid(grid)
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)

        # Plot faint bootstrapped curves
        num_bootstrap_samples = curves["_bootstrap_sample"].nunique() - 1
        for i in range(num_bootstrap_samples):
            ax.plot(
                curves.loc[lambda x: x["_bootstrap_sample"] == i, x_col],
                curves.loc[lambda x: x["_bootstrap_sample"] == i, y_col],
                alpha=alpha,
                color=bootstrap_color,
                linewidth=1,
            )

        # Plot main colored curve (scatter plot) with color bar
        fig, ax = self._make_plot(
            curves=curves,
            x_col=x_col,
            y_col=y_col,
            cmap=cmap,
            dpi=dpi,
            color_by=color_by,
            cbar_rng=cbar_rng,
            cbar_label=cbar_label,
            grid=grid,
            fig=fig,
            ax=ax,
        )

        return fig, ax

    def _get_metrics(
        self,
        imputed: bool = False,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """A helper function to get the metrics curves and scalars DataFrames
        for plotting.
        """
        curves = self.metrics.curves_imputed if imputed else self.metrics.curves
        scalars = self.metrics.scalars_imputed if imputed else self.metrics.scalars
        return curves, scalars



import pytest

from clscurves.config import MetricsAliases


@pytest.mark.parametrize(
    "key",
    [
        "tp",
        "fp",
        "fn",
        "tn",
        "tp_w",
        "fp_w",
        "fn_w",
        "tn_w",
        "precision",
        "tpr",
        "fpr",
        "tpr_w",
        "fpr_w",
        "frac",
        "frac_w",
        "thresh",
    ],
)
def test_rpf_dict_keys(key: str) -> None:
    cbar_dict = MetricsAliases.cbar_dict
    assert key in cbar_dict



from typing import List, Optional, Tuple

from matplotlib import pyplot as plt

from clscurves.plotter.plotter import MetricsPlotter
from clscurves.utils import MetricsResult


class PRGPlotter(MetricsPlotter):
    """Plot the PRG (Precision-Recall-Gain) curve.

    "Precision-Recall-Gain" plots defined
    `here <https://papers.nips.cc/paper/2015/file/33e8075e9970de0cfea955afd464\
    4bb2-Paper.pdf>`_.

    A "weighted recall" variant of Recall Gain is not discussed in the paper,
    and may not have any theoretical guarantees or reasonable interpretations.
    Because it's unclear how a positive-example weight should affect the
    quantities in a PRG plot, we do not support weighted plotting at this time.
    """

    def __init__(
        self,
        metrics: MetricsResult,
        score_is_probability: bool,
    ) -> None:
        super().__init__(metrics, score_is_probability)

    def plot_prg(
        self,
        title: str = "Precision-Recall-Gain Curve",
        cmap: str = "rainbow",
        color_by: str = "thresh",
        cbar_rng: Optional[List[float]] = None,
        cbar_label: Optional[str] = None,
        grid: bool = True,
        dpi: Optional[int] = None,
        bootstrapped: bool = False,
        bootstrap_alpha: float = 0.15,
        bootstrap_color: str = "black",
        imputed: bool = False,
        op_value: Optional[float] = None,
        return_fig: bool = False,
    ) -> Optional[Tuple[plt.Figure, plt.Axes]]:
        """Plot the PRG (Precision-Recall-Gain) curve.

        Parameters
        ----------
        title
            Title of plot.
        cmap
            Colormap string specification.
        color_by
            Name of key in metrics.curves that specifies which values to use
            when coloring points along the PRG curve; this should be either
            "frac" for fraction of cases flagged or "thresh" for score
            discrimination threshold.
        cbar_rng
            Specify a color bar range of the form [min_value, max_value] to
            override the default range.
        cbar_label
            Custom label to apply to the color bar. If None is supplied,
            the default ("Threshold Value" or "Fraction Flagged", depending on
            the ``color_by`` value) will be used.
        grid
            Whether to plot grid lines.
        dpi
            Resolution in "dots per inch" of resulting figure. If not
            specified, the Matplotlib default will be used. A good rule of
            thumb is 150 for good quality at normal screen resolutions and 300
            for high quality that maintains sharp features after zooming in.
        bootstrapped
            Specifies whether bootstrapped curves should be plotted behind the
            main colored performance scatter plot.
        bootstrap_alpha
            Opacity of bootstrap curves.
        bootstrap_color
            Color of bootstrap curves.
        imputed
            Whether to plot imputed curves.
        op_value
            Threshold value to plot a confidence ellipse for when the plot is
            bootstrapped.
        return_fig
            If set to True, will return (fig, ax) as a tuple instead of
            plotting the figure.
        """

        # Get metrics
        curves, scalars = self._get_metrics(imputed=imputed)

        # Specify which values to plot in X and Y
        x_col = "recall_gain"
        y_col = "precision_gain"

        # Make plot
        if not bootstrapped:
            fig, ax = self._make_plot(
                curves=curves,
                x_col=x_col,
                y_col=y_col,
                cmap=cmap,
                dpi=dpi,
                color_by=color_by,
                cbar_rng=cbar_rng,
                cbar_label=cbar_label,
                grid=grid,
            )
        else:
            fig, ax = self._make_bootstrap_plot(
                curves=curves,
                x_col=x_col,
                y_col=y_col,
                cmap=cmap,
                dpi=dpi,
                color_by=color_by,
                cbar_rng=cbar_rng,
                cbar_label=cbar_label,
                grid=grid,
                alpha=bootstrap_alpha,
                bootstrap_color=bootstrap_color,
            )

        # Extract PRG AUC
        scalars = scalars.loc[lambda x: x["_bootstrap_sample"].isnull()]
        auc = scalars["prg_auc"].iloc[0]

        # Extract class imbalance
        imb = scalars["imbalance"].iloc[0]

        # Compute the ratio of class sizes
        class_ratio = 1 / (imb + 1e-9) - 1

        # Plot line of randomness
        ax.plot([0, 1], [1, 0], "k-")

        # Add AUC to plot
        ax.text(
            x=0.06,
            y=0.13,
            s="%sAUPRG = %.3f" % ("Mean " if bootstrapped else "", auc),
            ha="left",
            va="center",
            bbox=dict(facecolor="gray", alpha=0.1, boxstyle="round"),
        )

        # Add class imbalance to plot
        ax.text(
            x=0.06,
            y=0.07,
            s="%sClass Imb. = %.1f : 1"
            % ("Mean " if bootstrapped else "", class_ratio),
            ha="left",
            va="center",
            bbox=dict(facecolor="gray", alpha=0.1, boxstyle="round"),
        )

        # Plot 95% confidence ellipse
        if op_value is not None:
            self._add_op_ellipse(
                curves=curves,
                op_value=op_value,
                x_col=x_col,
                y_col=y_col,
                ax=ax,
                thresh_key=color_by,
            )

        # Set labels
        ax.set_xlabel("Recall Gain = (Rec - pp)/((1 - pp) * Rec)")
        ax.set_ylabel("Precision Gain = (Prec - pp)/((1 - pp) * Prec)")
        ax.set_title(title)

        if return_fig:
            return fig, ax

        else:
            # Display and close plot
            plt.show()
            plt.close()

        return None



from typing import List, Optional, Tuple

from matplotlib import pyplot as plt
from typing_extensions import Literal

from clscurves.plotter.plotter import MetricsPlotter
from clscurves.utils import MetricsResult

Label = Literal["all", 0, 1, None]


class DistPlotter(MetricsPlotter):
    def __init__(
        self,
        metrics: MetricsResult,
        score_is_probability: bool,
        reverse_thresh: bool,
    ) -> None:
        super().__init__(metrics, score_is_probability)
        self.reverse_thresh = reverse_thresh

    def plot_dist(  # noqa: C901
        self,
        weighted: bool = False,
        label: Label = "all",  # noqa
        kind: str = "CDF",
        kernel_size: float = 10,
        log_scale: bool = False,
        title: Optional[str] = None,
        cmap: str = "rainbow",
        color_by: str = "recall",
        cbar_rng: Optional[List[float]] = None,
        cbar_label: Optional[str] = None,
        grid: bool = True,
        x_rng: Optional[List[float]] = None,
        y_rng: Optional[List[float]] = None,
        dpi: Optional[int] = None,
        bootstrapped: bool = False,
        bootstrap_alpha: float = 0.15,
        bootstrap_color: str = "black",
        imputed: bool = False,
        return_fig: bool = False,
    ) -> Optional[Tuple[plt.Figure, plt.Axes]]:
        """Plot the data distribution.

        This plots either the CDF (Cumulative Distribution Function) or PDF
        (Probability Density Function) curve.

        Parameters
        ----------
        weighted
            Specifies whether the weighted or unweighted fraction flagged
            should be used when computing the CDF or PDF. If unweighted, the
            fraction flagged is the number of cases flagged divided by the
            number of cases total. If weighted, it is the sum of the weights of
            all the cases flagged, divided by the sum of the weights of all
            the cases.
        label
            Class label to plot the CDF for; one of "all", 1, 0, or `None`.
        kind
            Either "cdf" or "pdf".
        kernel_size
            Used for PDF only: standard deviation of the Gaussian of kernel to
            use when smoothing the PDF curve.
        log_scale
            Boolean to specify whether the x-axis should be log-scaled.
        title
            Title of plot.
        cmap
            Colormap string specification.
        color_by
            Name of key in metrics.curves that specifies which values to use
            when coloring points along the PDF or CDF curve.
        cbar_rng
            Specify a color bar range of the form [min_value, max_value] to
            override the default range.
        cbar_label
            Custom label to apply to the color bar. If `None` is supplied, a
            default will be selected from the ``cbar_dict``.
        grid
            Whether to plot grid lines.
        x_rng
            Range of the horizontal axis.
        y_rng
            Range of the vertical axis.
        dpi
            Resolution in "dots per inch" of resulting figure. If not
            specified, the Matplotlib default will be used. A good rule of
            thumb is 150 for good quality at normal screen resolutions and 300
            for high quality that maintains sharp features after zooming in.
        bootstrapped
            Specifies whether bootstrapped curves should be plotted behind the
            main colored performance scatter plot.
        bootstrap_alpha
            Opacity of bootstrap curves.
        bootstrap_color
            Color of bootstrap curves.
        imputed
            Whether to plot imputed curves.
        return_fig
            If set to True, will return (fig, ax) as a tuple instead of
            plotting the figure.

        Returns
        -------
        Optional[Tuple[plt.Figure, plt.Axes]]
            The plot's figure and axis object.
        """
        assert label in ["all", 0, 1, None], '`label` must be in ["all", 0, 1, None]'

        kind = kind.lower()
        assert kind in ["cdf", "pdf"], '`kind` must be "cdf" or "pdf"'

        # Get metrics
        curves, _ = self._get_metrics(imputed=imputed)

        # Compute CDF
        _w = "_w" if weighted else ""
        if label == "all":
            cdf = 1 - curves["frac" + _w]
        else:
            raise NotImplementedError("TODO: Implement label=None case.")

        # Account for reversed-behavior thresholds
        if self.reverse_thresh:
            cdf = 1 - cdf

        curves["_cdf"] = cdf
        x_col = "thresh"
        y_col = "_cdf"

        # TODO: Support PDF via KDE.

        # Make plot
        if not bootstrapped:
            fig, ax = self._make_plot(
                curves=curves,
                x_col=x_col,
                y_col=y_col,
                cmap=cmap,
                dpi=dpi,
                color_by=color_by,
                cbar_rng=cbar_rng,
                cbar_label=cbar_label,
                grid=grid,
            )
        else:
            fig, ax = self._make_bootstrap_plot(
                curves=curves,
                x_col=x_col,
                y_col=y_col,
                cmap=cmap,
                dpi=dpi,
                color_by=color_by,
                cbar_rng=cbar_rng,
                cbar_label=cbar_label,
                grid=grid,
                alpha=bootstrap_alpha,
                bootstrap_color=bootstrap_color,
            )

        # Change x-axis range
        if x_rng:
            ax.set_xlim(*x_rng)

        # Log scale x-axis
        if log_scale:
            ax.set_xscale("log")
            if self.score_is_probability:
                x_rng = [0, 1] if x_rng is None else x_rng
                ax.set_xlim(*x_rng)

        # Change y-axis range
        if y_rng:
            ax.set_ylim(*y_rng)

        # Set aspect ratio
        x_size = x_rng[1] - x_rng[0] if x_rng else 1
        y_size = y_rng[1] - y_rng[0] if y_rng else 1
        ax.set_aspect(x_size / y_size)

        # Set labels
        weight_string = "Weighted " if weighted else ""
        label_string = f": Label = {label}" if label in [0, 1, None] else ""
        default_title = (
            f"{weight_string}CDF{label_string}"
            if kind == "cdf"
            else f"{weight_string}PDF{label_string}"
        )
        title = default_title if title is None else title
        ax.set_xlabel("Score")
        ax.set_ylabel("Cumulative Distribution" if kind == "cdf" else "Density")
        ax.set_title(title)

        if return_fig:
            return fig, ax

        else:
            # Display and close plot
            plt.show()
            plt.close()

        return None

    def plot_pdf(self, **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
        return self.plot_dist(kind="pdf", **kwargs)

    def plot_cdf(self, **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
        return self.plot_dist(kind="cdf", **kwargs)



import itertools
import logging
from multiprocessing import Pool
from typing import Optional

import numpy as np
import pandas as pd
import psutil
from numpy.random import default_rng
from scipy.integrate import trapz
from tqdm import tqdm
from typing_extensions import Literal

from clscurves.config import MetricsAliases
from clscurves.plotter.cost import CostPlotter
from clscurves.plotter.dist import DistPlotter
from clscurves.plotter.pr import PRPlotter
from clscurves.plotter.prg import PRGPlotter
from clscurves.plotter.rf import RFPlotter
from clscurves.plotter.roc import ROCPlotter
from clscurves.utils import MetricsResult

LOG = logging.getLogger(__name__)

NullFillMethod = Literal["0", "1", "imb", "prob"]


class MetricsGenerator(
    ROCPlotter,
    PRPlotter,
    PRGPlotter,
    RFPlotter,
    CostPlotter,
    DistPlotter,
    MetricsAliases,
):
    """A class to generate classification curve metrics.

    A class for computing Precision/Recall/Fraction metrics across a binary
    classification algorithm's full range of discrimination thresholds, and
    plotting those metrics as ROC (Receiver Operating Characteristic), PR
    (Precision & Recall), or RF (Recall & Fraction) plots. The input data
    format for this class is a PySpark DataFrame with at least a column of
    labels and a column of scores (with an optional additional column of label
    weights).
    """

    def __init__(
        self,
        predictions_df: Optional[pd.DataFrame] = None,
        max_num_examples: int = 100000,
        label_column: str = "label",
        score_column: str = "probability",
        weight_column: Optional[str] = None,
        score_is_probability: bool = True,
        reverse_thresh: bool = False,
        num_bootstrap_samples: int = 0,
        imbalance_multiplier: float = 1,
        null_prob_column: Optional[str] = None,
        null_fill_method: Optional[NullFillMethod] = None,
        seed: Optional[int] = None,
    ) -> None:
        """Instantiating this class computes all the metrics.

        Parameters
        ----------
        predictions_df : pd.DataFrame
            Input DataFrame which must contain a column of labels (integer
            column of 1s and 0s) and a column of scores (either a dense vector
            column with two elements [prob_0, prob_1] or a real-valued column
            of scores).
        max_num_examples : int
            Max number of rows to sample to prevent numpy memory limits from
            being exceeded.
        label_column : str
            Name of the column containing the example labels, which must all be
            either 0 or 1.
        score_column : str
            Name of the column containing the example scores which rank model
            predictions. Even though binary classification models typically
            output probabilities, these scores need not be bounded by 0 and 1;
            they can be any real value. This column can be either a real value
            numeric type or a 2-element vector of probabilities, with the first
            element being the probability that the example is of class 0 and
            the second element that the example is of class 1.
        weight_column : Optional[str]
            Name of the column containing label weights associated with each
            example. These weights are useful when the cost of classifying an
            example incorrectly varies from example to example; see fraud, for
            instance: getting high dollar value cases wrong is more costly than
            getting low dollar value cases wrong, so a good measure of recall
            is, "How much money did we catch", not "How many cases did we
            catch". If no column name is specified, all weights will be set to
            1.
        score_is_probability : bool
            Specifies whether the values in the score column are bounded by 0
            and 1. This controls how the threshold range is determined. If
            true, the threshold range will sweep from 0 to 1. If false, it will
            sweep from the minimum to maximum score value.
        num_bootstrap_samples : int
            Number of bootstrap samples to generate from the original data when
            computing performance metrics.
        reverse_thresh : bool
            Boolean indicating whether the score threshold should be treated as
            a lower bound on "positive" predictions (as is standard) or instead
            as an upper bound. If `True`, the threshold behavior will be
            reversed from standard so that any prediction falling BELOW a score
            threshold will be marked as positive, with all those falling above
            the threshold marked as negative.
        imbalance_multiplier : float
            Positive value to artifically increase the positive class example
            count by a multiplicative weighting factor. Use this if you're
            generating metrics for a data distribution with a class imbalance
            that doesn't represent the true distribution in the wild. For
            example, if you trained on a 1:1 artifically balanced data set, but
            you have a 10:1 class imbalance in the wild (i.e. 10 negative
            examples for every 1 positive example), set the
            ``imbalance_multiplier`` value to 10.
        null_prob_column : Optional[str]
            Column containing calibrated label probabilities to use as the
            sampling distribution for imputing null label values. We provide
            this argument so that you can evaluate a possibly-uncalibrated
            model score (specified by the `score_column` argument) on a
            different provided calibrated label distribution. If this argument
            is `None`, then the ``score_column`` will be used as the estimated
            label distribution when necessary.
        null_fill_method : Optional[NullFillMethod]
            Methods to use when filling in null label values. Possible values:
                * "0" - fill with 0
                * "1" - fill with 1
                * "imb" - fill randomly according to the class imbalance of
                    labeled examples
                * "prob" - fill randomly according to the ``score_column``
                    probability distribution or the ``null_prob_column``
                    probability distribution, if provided.
            If a method is provided, once the default metrics DF is computed
            without imputing any null labels, then a new metrics DF will be
            computed for each method and stored in a ``curves_imputed``
            object. If not, only the default metrics DF will be computed.
        seed : Optional[int]
            Random seed for bootstrapping.

        Examples
        --------
        >>> mg = MetricsGenerator(
                predictions_df,
                label_column="label",
                score_column="score",
                weight_column="weight",
                score_is_probability=False,
                reverse_thresh=False,
                num_bootstrap_samples=20,
                seed=123,
            )

        >>> mg.plot_pr(bootstrapped=True)
        >>> mg.plot_roc()
        """

        self.predictions_df = predictions_df
        self.max_num_examples = max_num_examples
        self.label_column = label_column
        self.score_column = score_column
        self.weight_column = weight_column
        self.score_is_probability = score_is_probability
        self.reverse_thresh = reverse_thresh
        self.num_bootstrap_samples = num_bootstrap_samples
        self.imbalance_multiplier = imbalance_multiplier
        self.null_prob_column = null_prob_column
        self.null_fill_method = null_fill_method
        self.null_probabilities = None
        self.seed = seed

        # Metrics to be populated
        self.metrics: MetricsResult

        # Print imbalance multiplier warning
        if self.imbalance_multiplier != 1:
            print(f"Artificial imbalance multiplier: {self.imbalance_multiplier}")

        if null_fill_method not in [None, "0", "1", "imb", "prob"]:
            raise ValueError(
                f"Invalid null_fill_method: {null_fill_method}. Must be one of "
                f"None, '0', '1', 'imb', or 'prob'."
            )

        if predictions_df is not None:
            self.compute_all_metrics(predictions_df)

    def _get_rng(self, idx: int) -> np.random.Generator:
        """Get random generator for bootstrap sampling."""
        seed = None if self.seed is None else self.seed + idx + 1
        return default_rng(seed)

    def compute_all_metrics(
        self,
        predictions_df: pd.DataFrame,
        return_results: bool = False,
    ) -> Optional[MetricsResult]:
        """Compute all metrics."""
        print("Computing metrics...")

        # Check if there are any null labels
        labels_contain_null = predictions_df[self.label_column].isnull().any()
        if labels_contain_null:
            LOG.warning(" >>> WARNING: Labels contain null values.")

        # Keep only relevant columns
        cols = [self.label_column, self.score_column]
        if self.weight_column:
            cols.append(self.weight_column)
        df_sample = predictions_df[cols].copy()

        # Sample input data if too large
        if len(df_sample) > self.max_num_examples:
            rng = self._get_rng(-1)
            seed = int(rng.random() * 2**32)
            df_sample = df_sample.sample(
                n=self.max_num_examples,
                random_state=seed,
            )

        # List configurations to compute
        bootstrap_sample_options = [None, *range(self.num_bootstrap_samples)]
        null_fill_methods = list(set([None, self.null_fill_method]))
        all_options = itertools.product(
            bootstrap_sample_options,
            null_fill_methods,
        )

        # Compute metrics
        curves = pd.DataFrame()
        scalars = pd.DataFrame()
        num_workers = psutil.cpu_count()
        with Pool(num_workers) as pool:
            for metrics in tqdm(
                pool.starmap(
                    self.compute_metrics,
                    [
                        (df_sample, *options, self._get_rng(i))
                        for i, options in enumerate(all_options)
                    ],
                )
            ):
                curves = pd.concat([curves, metrics.curves])
                scalars = pd.concat([scalars, metrics.scalars])

        # Separate imputed metrics from default metrics
        curves_default = curves.loc[curves["_null_fill_method"].isnull()]
        scalars_default = scalars.loc[scalars["_null_fill_method"].isnull()]
        curves_imputed = curves.loc[curves["_null_fill_method"].notnull()]
        scalars_imputed = scalars.loc[scalars["_null_fill_method"].notnull()]

        metrics = MetricsResult(
            curves=curves_default,
            scalars=scalars_default,
            curves_imputed=curves_imputed,
            scalars_imputed=scalars_imputed,
        )

        print("Metrics computation complete.")

        # Optional return
        if return_results:
            return metrics

        self.metrics = metrics

        return None

    def compute_metrics(
        self,
        predictions_df: pd.DataFrame,
        bootstrap_sample: Optional[int] = None,
        null_fill_method: Optional[NullFillMethod] = None,
        rng: np.random.Generator = default_rng(),
    ) -> MetricsResult:
        """Compute metrics for a single bootstrap sample."""

        # Keep or drop nulls
        if null_fill_method is None:
            _df = predictions_df.dropna(subset=[self.label_column])
        else:
            _df = predictions_df

        # Make a bootstrap
        if bootstrap_sample is not None:
            _df = self._make_bootstrap(_df, rng)

        # Impute null labels
        labels = _df[self.label_column].values
        if null_fill_method is not None:
            labels = self._fill_null_labels(
                labels=labels,
                null_fill_method=null_fill_method,
                null_probs=self.null_probabilities,
                rng=rng,
            )
            assert not np.isnan(labels).any()

        # Compute metrics
        metrics = self._compute_metrics(
            scores=_df[self.score_column].values,
            labels=labels,
            weights=_df[self.weight_column].values if self.weight_column else None,
            reverse_thresh=self.reverse_thresh,
        )

        # Attach metadata to output
        metrics.curves["_bootstrap_sample"] = bootstrap_sample
        metrics.curves["_null_fill_method"] = null_fill_method
        metrics.scalars["_bootstrap_sample"] = bootstrap_sample
        metrics.scalars["_null_fill_method"] = null_fill_method

        return metrics

    def _compute_metrics(
        self,
        scores: np.ndarray,
        labels: np.ndarray,
        weights: Optional[np.ndarray] = None,
        reverse_thresh: bool = False,
    ) -> MetricsResult:
        """Compute metrics.

        Parameters
        ----------
        scores : np.ndarray
            Array of scores.
        labels : np.ndarray
            Binary array of labels. If labels are real-valued, then all values
            > 0 will be treated as positive examples, and all values <= 0 will
            be treated as negative examples.
        weights : Optional[np.ndarray]
            Array of weights associated with each example.
        reverse_thresh : bool
            Boolean indicating whether the score threshold should be treated as
            a lower bound on "positive" predictions (as is standard) or instead
            as an upper bound. If ``True``, the threshold behavior will be
            reversed from standard so that any prediction falling BELOW a score
            threshold will be marked as positive, with all those falling above
            the threshold marked as negative.

        Returns
        -------
        MetricsResult
            A class containing the computed metrics.
        """
        if np.isnan(labels).any():
            raise ValueError("Labels contain null values.")

        # Put arrays into DataFrame, treating scores as thresholds
        df = pd.DataFrame(
            {
                "thresh": scores,
                "label": (labels > 0).astype(int),
                "weight": weights if weights is not None else np.ones(len(scores)),
            },
        )
        df["weight_pos"] = df["label"] * df["weight"]

        # Collapse identical threshold values, and sort
        df = (
            df.groupby("thresh")
            .agg(
                label=("label", "sum"),
                weight=("weight", "sum"),
                weight_pos=("weight_pos", "sum"),
                num=("label", "count"),
            )
            .reset_index()
        )
        df = df.sort_values("thresh", ascending=not reverse_thresh)

        # Compute scalar values
        _num_examples = df["num"].sum()
        _num_examples_pos = df["label"].sum()
        num_examples_pos = _num_examples_pos * self.imbalance_multiplier
        num_examples_neg = _num_examples - _num_examples_pos
        num_examples = num_examples_pos + num_examples_neg
        _tot_weight = df["weight"].sum()
        _tot_weight_pos = df["weight_pos"].sum()
        tot_weight_pos = _tot_weight_pos * self.imbalance_multiplier
        tot_weight_neg = _tot_weight - _tot_weight_pos
        tot_weight = tot_weight_pos + tot_weight_neg
        imbalance = num_examples_pos / (num_examples_pos + num_examples_neg)

        # Add extra threshold value
        epsilon = 1e-6
        multiplier = 1 if reverse_thresh else -1
        extra_thresh = df.iloc[0]["thresh"] + multiplier * epsilon
        extra_row = pd.DataFrame([[extra_thresh, 0, 0, 0, 0]], columns=df.columns)
        df = pd.concat([extra_row, df]).reset_index(drop=True)

        # Compute confusion matrix
        df["pred_neg"] = df["num"].cumsum()
        df["pred_pos"] = _num_examples - df["pred_neg"]
        df["fn"] = df["label"].cumsum()
        df["tn"] = df["pred_neg"] - df["fn"]
        df["tp"] = _num_examples_pos - df["fn"]
        df["fp"] = _num_examples - df["pred_neg"] - df["tp"]
        df["pred_pos"] *= self.imbalance_multiplier
        df["fn"] *= self.imbalance_multiplier
        df["tp"] *= self.imbalance_multiplier

        df["recall"] = df["tp"] / (df["tp"] + df["fn"])
        df["precision"] = df["tp"] / (df["tp"] + df["fp"])
        df["frac"] = df["pred_pos"] / num_examples
        df["f1"] = 2 * df["precision"] * df["recall"] / (df["precision"] + df["recall"])
        df["fpr"] = df["fp"] / num_examples_neg
        df["fdr"] = df["fp"] / (df["fp"] + df["tp"])
        df["recall_gain"] = self._compute_gain(df["recall"], imbalance)
        df["precision_gain"] = self._compute_gain(df["precision"], imbalance)

        # Compute weighted confusion matrix
        df["pred_neg_w"] = df["weight"].cumsum()
        df["pred_pos_w"] = _tot_weight - df["pred_neg_w"]
        df["fn_w"] = df["weight_pos"].cumsum()
        df["tn_w"] = df["pred_neg_w"] - df["fn_w"]
        df["tp_w"] = _tot_weight_pos - df["fn_w"]
        df["fp_w"] = _tot_weight - df["pred_neg_w"] - df["tp_w"]
        df["pred_pos_w"] *= self.imbalance_multiplier
        df["fn_w"] *= self.imbalance_multiplier
        df["tp_w"] *= self.imbalance_multiplier

        df["recall_w"] = df["tp_w"] / (df["tp_w"] + df["fn_w"])
        df["precision_w"] = df["tp_w"] / (df["tp_w"] + df["fp_w"])
        df["frac_w"] = df["pred_pos_w"] / tot_weight
        df["fpr_w"] = df["fp_w"] / tot_weight_neg
        df["fdr_w"] = df["fp_w"] / (df["fp_w"] + df["tp_w"])

        # Fill nulls
        df.fillna({col: 0 for col in df.columns if col != "label"}, inplace=True)

        # Scalars
        scalars = {
            "num_examples": num_examples,
            "num_examples_pos": num_examples_pos,
            "num_examples_neg": num_examples_neg,
            "tot_weight": tot_weight,
            "tot_weight_pos": tot_weight_pos,
            "tot_weight_neg": tot_weight_neg,
            "imbalance": imbalance,
            "roc_auc": np.abs(trapz(df["recall"], df["fpr"])),
            "pr_auc": np.abs(trapz(df["precision"], df["recall"])),
            "rf_auc": np.abs(trapz(df["recall"], df["frac"])),
            "roc_auc_w": np.abs(trapz(df["recall_w"], df["fpr_w"])),
            "pr_auc_w": np.abs(trapz(df["precision_w"], df["recall_w"])),
            "rf_auc_w": np.abs(trapz(df["recall_w"], df["frac_w"])),
            "prg_auc": np.abs(trapz(df["precision_gain"], df["recall_gain"])),
        }
        scalars = pd.DataFrame([scalars])

        return MetricsResult(curves=df, scalars=scalars)

    def _make_bootstrap(
        self,
        df: pd.DataFrame,
        rng: np.random.Generator = default_rng(),
    ) -> pd.DataFrame:
        """Make bootstrap sample.

        Sample with replacement from the input DataFrame to create a bootstrap
        sample of the same size as the input DataFrame.
        """
        random_state = (
            np.random.RandomState(int(rng.random() * 2**32)) if rng else None
        )
        return df.sample(
            n=len(df),
            replace=True,
            random_state=random_state,
        )

    def _fill_null_labels(
        self,
        labels: np.ndarray,
        null_fill_method: NullFillMethod,
        null_probs: Optional[np.ndarray] = None,
        rng: np.random.Generator = default_rng(),
    ) -> np.ndarray:
        """Fill null labels according to the specified method.

        Parameters
        ----------
        labels : np.ndarray
            Array of labels, some of which may be null.
        null_fill_method : NullFillMethod
            Method to use when filling null labels.
            * "0" -- fill unknown labels with 0.
            * "1" -- fill unknown labels with 1.
            * "imb" -- fill unknown labels with 0 or 1 probabilistically
                according to the class imbalance of the known labels.
            * "prob" -- fill unknown labels with 0 or 1 probabilistically
                according to the probability-calibrated model score.
        null_probs : Optional[np.ndarray]
            Array of probabilities to use when filling null labels. Only
            required if ``null_fill_method`` is "prob".
        """
        labels = labels.astype(float)
        imbalance = (labels > 0).sum() / len(labels)
        if null_fill_method == "0":
            return np.where(np.isnan(labels), 0, labels)
        if null_fill_method == "1":
            return np.where(np.isnan(labels), 1, labels)
        if null_fill_method == "imb":
            labels_from_imb = (rng.random(*labels.shape) < imbalance).astype(int)
            return np.where(np.isnan(labels), labels_from_imb, labels)
        if null_fill_method == "prob":
            if null_probs is None:
                raise ValueError(
                    "Must provide null_probs when using null_fill_method='prob'."
                )
            return np.where(
                np.isnan(labels),
                (rng.random(*labels.shape) < null_probs).astype(int),
                labels,
            )

    @staticmethod
    def _compute_gain(
        metric: np.ndarray,
        imbalance: float,
    ) -> np.ndarray:
        """Compute "gain".

        As defined in the "Precision-Recall-Gain" paper
        `here <https://papers.nips.cc/paper/2015/file/33e8075e9970de0cfea955afd464\
        4bb2-Paper.pdf>`_.
        """
        return np.clip(
            (metric - imbalance) / ((1 - imbalance) * metric),
            a_min=0,
            a_max=1,
        )



from dataclasses import dataclass
from typing import Optional

import pandas as pd


@dataclass
class MetricsResult:
    """A class to hold the results of a metrics computation.

    Parameters
    ----------
    curves : pd.DataFrame
        DataFrame containing metrics at each threshold.
    scalars : pd.DataFrame
        DataFrame containing scalar metrics which apply to the entire dataset.
    curves_imputed : Optional[pd.DataFrame]
        DataFrame containing metrics at each threshold after imputation.
    scalars_imputed : Optional[pd.DataFrame]
        DataFrame containing scalar metrics which apply to the entire dataset
        after imputation.
    """

    curves: pd.DataFrame
    scalars: pd.DataFrame
    curves_imputed: Optional[pd.DataFrame] = None
    scalars_imputed: Optional[pd.DataFrame] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clscurves', 'clscurves.plotter', 'clscurves.tests']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.8.2,<4.0.0',
 'numpy>=1.20.3,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'psutil>=5.9.5,<6.0.0',
 'scipy>=1.6.3,<2.0.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'clscurves',
    'version': '0.3.0',
    'description': 'Compute and plot bootstrapped performance curves for classification problems.',
    'long_description': '# classification-curves\n\nA library for computing and plotting bootstrapped metrics (ROC curves,\nPrecision-Recall curves, etc.) to evaluate the performance of a classification\nmodel.\n\n## Example\n```python\nmg = MetricsGenerator(\n    predictions_df,\n    label_column = "label",\n    score_column = "score",\n    weight_column = "weight",\n    score_is_probability = False,\n    reverse_thresh = False,\n    num_bootstrap_samples = 20)\n\nmg.plot_pr(bootstrapped = True)\nmg.plot_roc()\n```\n        ',
    'author': 'Christopher Bryant',
    'author_email': 'cbryant@berkeley.edu',
    'maintainer': 'Christopher Bryant',
    'maintainer_email': 'cbryant@berkeley.edu',
    'url': 'https://github.com/chrismbryant/classification-curves',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10.0,<3.11',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd

from .. import MetricsGenerator


def test_compute_confusion_matrix() -> None:
    scores = np.array([1, 2, 2, 2, 4, 5, 6, 8, 9, 10])
    labels = np.array([0, 0, 1, 1, 0, 1, 1, 0, 5, np.nan])
    weights = np.array([2, 2, 2, 2, 2, 1, 1, 1, 1, 1])
    df = pd.DataFrame(
        {
            "scores": scores,
            "labels": labels,
            "weights": weights,
        }
    )

    mg = MetricsGenerator(
        df,
        label_column="labels",
        score_column="scores",
        weight_column="weights",
        num_bootstrap_samples=0,
        seed=123,
    )

    # TODO: Make a real test
    assert mg is not None
